<?php

declare(strict_types=1);

namespace mirzaev\arming_bot\models;

// Files of the project
use mirzaev\arming_bot\models\core,
	mirzaev\arming_bot\models\product,
	mirzaev\arming_bot\models\category,
	mirzaev\arming_bot\models\entry,
	mirzaev\arming_bot\models\traits\files,
	mirzaev\arming_bot\models\traits\yandex\disk as yandex;

// Framework for ArangoDB
use mirzaev\arangodb\collection,
	mirzaev\arangodb\document;

// Framework for Excel
use avadim\FastExcelReader\Excel as excel;

// Built-in libraries
use exception;

/**
 * Model of the catalog
 *
 * @package mirzaev\arming_bot\models
 *
 * @license http://www.wtfpl.net/ Do What The Fuck You Want To Public License
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
final class catalog extends core
{
	use yandex, files {
		yandex::download as yandex;
	}

	/**
	 * Collect parameter from all products
	 *
	 * @param string $documment Path to the EXCEL-document
   * @param int &$categories_loaded Counter of loaded categories
   * @param int &$categories_created Counter of created categories
   * @param int &$categories_updated Counter of updated categories
   * @param int &$categories_deleted Counter of deleted categories
   * @param int &$categories_new Counter of new categories
   * @param int &$categories_old Counter of old categories
   * @param int &$products_loaded Counter of loaded products
   * @param int &$products_created Counter of created products
   * @param int &$products_updated Counter of updated products
   * @param int &$products_deleted Counter of deleted products
   * @param int &$products_new Counter of new products
   * @param int &$products_old Counter of old products
   * @param string &$language Language (en, ru...)
	 * @param array &$errors Registry of errors
	 *
	 * @return void
	 *
	 * @todo 
	 * 1. Сначала создать все категории и затем снова по циклу пройтись уже создавать entry между ними
   * 2. Сжимать изображения
	 */
	public static function import(
		string $document,
		int &$categories_loaded = 0,
		int &$categories_created = 0,
		int &$categories_updated = 0,
		int &$categories_deleted = 0,
		int &$categories_old = 0,
		int &$categories_new = 0,
		int &$products_loaded = 0,
		int &$products_created = 0,
		int &$products_updated = 0,
		int &$products_deleted = 0,
		int &$products_old = 0,
		int &$products_new = 0,
		string $language = 'en',
		array &$errors = []
	): void {
		try {
			// Initializing the spreadsheet
			$spreadsheet = excel::open($document);

			// Inititalizing worksheets
			$categories = $spreadsheet->getSheet('Категории');
			$products = $spreadsheet->getSheet('Товары');

			// Counting old documents
			$categories_old = collection::count(category::COLLECTION, errors: $errors);
			$products_old = collection::count(product::COLLECTION, errors: $errors);

			// Initializing the buffer of handler categories and products
			$handled = [
				'categories' => [],
				'products' => []
			];

			foreach (
				$categories->nextRow(
					[
						'A' => 'identifier',
						'B' => 'name',
						'C' => 'category',
						'D' => 'images',
						'E' => 'position'
					],
					excel::KEYS_FIRST_ROW
				) as $number => $row
			) {
				// Iterate over categories

				try {
					if (!empty($row['identifier']) && !empty($row['name'])) {
						// All required cells are filled in

						// Incrementing the counter of loaded categories
						++$categories_loaded;

						// Declaring the variable with the status that a new category has been created
						$created = false;

						// Declaring the variable with the category
						$category = null;

						try {
							// Initializing the category
							$category = new category(document: category::_read('d.identifier == @identifier', parameters: ['identifier' => (int) $row['identifier']], errors: $errors)[0] ?? null);

							// Initializing name of the category
							$category->name[$language] === $row['name'] || $category->name = [[$language => $row['name']]] + $category->name ?? [];

							// Initializing position of the category
							$category->position === $row['position'] || $category->position = $row['position'];
						} catch (exception $e) {
							// Not found the category

							// Creating the category
							$_id = category::write((int) $row['identifier'], [$language => $row['name']], $row['position'] ?? null, $errors);

							// Initializing the category
							$category = new category(document: $created = category::_read('d._id == @_id', parameters: ['_id' => $_id], errors: $errors)[0] ?? null);

							// Incrementing the counter of created categories
							if ($created) ++$categories_created;
						};

						if ($category instanceof category) {
							// Found the category

							if (!empty($row['category'])) {
								// Received the ascendant category

								// Initializing the ascendant category
								$ascendant = new category(document: category::_read('d.identifier == @identifier', parameters: ['identifier' => (int) $row['category']], errors: $errors)[0] ?? null);

								if ($ascendant instanceof category) {
									// Found the ascendant category

									// Deleting entries of the category in ArangoDB
									entry::banish($category, $errors);

									// Writing the category as an entry to the ascendant category in ArangoDB
									entry::write($category, $ascendant, $errors);
								}
							}

							if (!empty($row['images'])) {
								// Received images

								// Initializing new images of the category
								$images = explode(' ', trim($row['images']));

								// Reinitialize images? (true, if no images found or their amount does not match)
								$reinitialize = !$category->images || count($category->images) !== count($images);

								// Checking the identity of existing images with new images (if reinitialization of images has not yet been requested)
								if (!$reinitialize)	foreach ($category->images as $key => $image) if ($reinitialize = $image['source'] !== $images[$key]) break;

								if ($reinitialize) {
									// Requested reinitialization of images

									// Initializing the buffer of images
									$buffer = [];

									foreach ($images as $index => $image) {
										// Iterating over new images

										// Skipping empty URI`s
										if (empty($image = trim($image))) continue;

										// Initializing path to directory of the image in storage
										$directory = DIRECTORY_SEPARATOR . 'categories' . DIRECTORY_SEPARATOR . $row['identifier'];

										// Initializing URL of the image in storage
										$url = STORAGE . $directory;

										// Initializing URN of the image in storage
										$urn = $index . '.jpg';

										// Initializing URI of the image in storage
										$uri = $url . DIRECTORY_SEPARATOR . $urn;

										// Initializing the directory in storage
										if (!file_exists($url))	mkdir($url, 0775, true);

										if (static::yandex($image, $uri, errors: $errors)) {
											// The image is downloaded

											// Writing the image to the buffer if images
											$buffer[] = [
												'source' => $image,
												'storage' => $directory . DIRECTORY_SEPARATOR . $urn
											];
										}
									}

									// Initializing images of the category
									$category->images = $buffer;
								}
							}

							// Writing in ArangoDB
							$updated = document::update($category->__document(), errors: $errors);

							// Incrementing the counter of updated categories
							if ($updated && !$created) ++$categories_updated;
						} else throw new exception("Failed to initialize category: {$row['name']} ($number)");
					}

					// Writing to the registry of handled categories and products
					$handled['categories'][] = $row['identifier'];
				} catch (exception $e) {
					// Writing to the registry of errors
					$errors[] = [
						'text' => $e->getMessage(),
						'file' => $e->getFile(),
						'line' => $e->getLine(),
						'stack' => $e->getTrace()
					];
				}
			}

			foreach (
				$products->nextRow(
					[
						'A' => 'identifier',
						'B' => 'name',
						'C' => 'category',
						'D' => 'description',
						'E' => 'cost',
						'F' => 'weight',
						'G' => 'x',
						'H' => 'y',
						'I' => 'z',
						'J' => 'brand',
						'K' => 'compatibility',
						'L' => 'images',
						'M' => 'position'
					],
					excel::KEYS_FIRST_ROW
				) as $number => $row
			) {
				// Iterate over products

				try {
					if (!empty($row['identifier']) && !empty($row['name'])) {
						// All required cells are filled in

						// Incrementing the counter of loaded products
						++$products_loaded;

						// Declaring the variable with the status that a new product has been created
						$created = false;

						// Declaring the variable with the product
						$product = null;

						try {
							// Initializing the product
							$product = new product(document: product::_read('d.identifier == %u', parameters: ['identifier' => (int) $row['identifier']], errors: $errors)[0] ?? null);

							// Initializing name of the category
							$product->name[$language] === $row['name'] || $product->name = [[$language => $row['name']]] + $product->name ?? [];

							// Initializing position of the product
							$product->position === $row['position'] || $product->position = $row['position'];
						} catch (exception $e) {
							// Not found the product

							// Creating the product
							$_id = product::write(
								(int) $row['identifier'],
								[$language => $row['name']],
								[$language => $row['description']],
								(float) $row['cost'],
								(float) $row['weight'],
								['x' => $row['x'], 'y' => $row['y'], 'z' => $row['z']],
								$row['position'] ?? null,
								errors: $errors
							);

							// Initializing the product
							$product = new product(document: $created = product::_read(sprintf('d._id == "%s"', $_id), errors: $errors)[0] ?? null);

							// Incrementing the counter of created products
							if ($created) ++$products_created;
						}

						if ($product instanceof product) {
							// Found the product

							if (!empty($row['category'])) {
								// Received the category

								// Initializing the category
								$category = new category(document: category::_read(sprintf('d.identifier == %u', (int) $row['category']), errors: $errors)[0] ?? null);

								if ($category instanceof category) {
									// Found the ascendant category

									// Deleting entries of the product in ArangoDB
									entry::banish($product, $errors);

									// Writing the product as an entry to the ascendant category in ArangoDB
									entry::write($product, $category, $errors);
								}
							}

							if (!empty($row['images'])) {
								// Received images

								// Initializing new images of the category
								$images = explode(' ', trim($row['images']));

								// Reinitialize images? (true, if no images found or their amount does not match)
								$reinitialize = !$product->images || count($product->images) !== count($images);

								// Checking the identity of existing images with new images (if reinitialization of images has not yet been requested)
								if (!$reinitialize)	foreach ($product->images as $key => $image) if ($reinitialize = $image['source'] !== $images[$key]) break;

								if ($reinitialize) {
									// Requested reinitialization of images

									// Initializing the buffer of images
									$buffer = [];

									foreach ($images as $index => $image) {
										// Iterating over new images

										// Skipping empty URI`s
										if (empty($image = trim($image))) continue;

										// Initializing path to directory of the image in storage
										$directory = DIRECTORY_SEPARATOR . 'products' . DIRECTORY_SEPARATOR . $row['identifier'];

										// Initializing URL of the image in storage
										$url = STORAGE . $directory;

										// Initializing URN of the image in storage
										$urn = $index . '.jpg';

										// Initializing URI of the image in storage
										$uri = $url . DIRECTORY_SEPARATOR . $urn;

										// Initializing the directory in storage
										if (!file_exists($url))	mkdir($url, 0775, true);

										if (static::yandex($image, $uri, errors: $errors)) {
											// The image is downloaded

											// Writing the image to the buffer if images
											$buffer[] = [
												'source' => $image,
												'storage' => $directory . DIRECTORY_SEPARATOR . $urn
											];
										}
									}

									// Initializing images of the category
									$product->images = $buffer;
								}
							}

							// Writing in ArangoDB
							$updated = document::update($product->__document(), errors: $errors);

							// Incrementing the counter of updated categories
							if ($updated && !$created) ++$products_updated;
						} else throw new exception("Failed to initialize product: {$row['name']} ($number)");
					}

					// Writing to the registry of handled categories and products
					$handled['products'][] = $row['identifier'];
				} catch (exception $e) {
					// Writing to the registry of errors
					$errors[] = [
						'text' => $e->getMessage(),
						'file' => $e->getFile(),
						'line' => $e->getLine(),
						'stack' => $e->getTrace()
					];
				}
			}

			// Deleting old categories
			foreach (
				category::_read(
					/* filter: sprintf('%s - d.updated > 3600', time()), */
					sort: 'd.updated DESC',
					amount: 100000,
					errors: $errors
				) as $document
			) {
				// Iterating over categories

				// Initializing the category
				$category = new category(document: $document);

				if (
					$category instanceof category
					&& array_search($category->identifier, $handled['categories']) === false
				) {
					// Not found identifier of the product in the buffer of handled categories and products

					// Deleting images of the category from storage
					static::delete(STORAGE . DIRECTORY_SEPARATOR . 'categories' . DIRECTORY_SEPARATOR . $category->identifier, errors: $errors);

					// Deleting entries of the category in ArangoDB
					entry::banish($category, errors: $errors);

					// Deleting the category in ArangoDB
					document::delete($category->__document(), errors: $errors);

					// Incrementing the counter of deleted categories
					++$categories_deleted;
				}
			}

			// Deleting old products
			foreach (
				product::_read(
					/* filter: sprintf('%s - d.updated > 3600', time()), */
					sort: 'd.updated DESC',
					amount: 100000,
					errors: $errors
				) as $document
			) {
				// Iterating over products

				// Initializing the category
				$product = new product(document: $document);

				if (
					$product instanceof product
					&& array_search($product->identifier, $handled['products']) === false
				) {
					// Not found identifier of the product in the buffer of handled categories and products

					// Deleting images of the product from storage
					static::delete(STORAGE . DIRECTORY_SEPARATOR . 'products' . DIRECTORY_SEPARATOR . $product->identifier, errors: $errors);

					// Deleting entries of the product in ArangoDB
					entry::banish($product, errors: $errors);

					// Deleting the product in ArangoDB
					document::delete($product->__document(), errors: $errors);

					// Incrementing the counter of deleted products
					++$products_deleted;
				}
			}

			// Counting new documents
			$categories_new = collection::count(category::COLLECTION, errors: $errors);
			$products_new = collection::count(product::COLLECTION, errors: $errors);
		} catch (exception $e) {
			// Writing to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}
	}
}
