<?php

declare(strict_types=1);

namespace mirzaev\arming_bot\models;

// Files of the project
use mirzaev\arming_bot\models\core,
	mirzaev\arming_bot\models\traits\document as arangodb_document_trait;

// Framework for ArangoDB
use mirzaev\arangodb\collection,
	mirzaev\arangodb\document;

// Built-in libraries
use exception;

/**
 * Model of a product
 *
 * @package mirzaev\arming_bot\models
 *
 * @license http://www.wtfpl.net/ Do What The Fuck You Want To Public License
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
final class product extends core
{
	use arangodb_document_trait;

	/**
	 * Name of the collection in ArangoDB
	 */
	final public const string COLLECTION = 'product';

	/**
	 * Write a product
	 *
	 * @param int $identifier Identifier (unique)
	 * @param array $name Name [['en' => value], ['ru' => значение]]
	 * @param array|null $description Description [['en' => value], ['ru' => значение]]
	 * @param float $cost Cost
	 * @param float $weight Weight
	 * @param array $dimensions Dimensions ['x' => 0.0, 'y' => 0.0, 'z' => 0.0]
	 * @param array $images Images (first will be thumbnail)
	 * @param int|null $position Position for sorting in the catalog (ASC)
	 * @param array $data Data
	 * @param array &$errors Registry of errors
	 *
	 * @return string|null Identifier (_id) of instance of the product document in ArangoDB, if created
	 *
	 * @todo
	 * 1. Bind parameters
	 */
	public static function write(
		int $identifier,
		array $name = [['en' => 'ERROR']],
		?array $description = [['en' => 'ERROR']],
		float $cost = 0,
		float $weight = 0,
		array $dimensions = ['x' => 0, 'y' => 0, 'z' => 0],
		array $images = [],
		?int $position = null,
		array $data = [],
		array &$errors = []
	): string|null {
		try {
			if (collection::initialize(static::COLLECTION, static::TYPE, errors: $errors)) {
				// Initialized the collection

				// Writing in ArangoDB and exit (success)
				return document::write(
					static::COLLECTION,
					[
						'identifier' => $identifier,
						'name' => $name,
						'description' => $description,
						'cost' => $cost ?? 0,
						'weight' => $weight ?? 0,
						'dimensions' => [
							'x' => $dimensions['x'] ?? 0,
							'y' => $dimensions['y'] ?? 0,
							'z' => $dimensions['z'] ?? 0,
						],
						'images' => $images,
						'position' => $position,
						'version' => ROBOT_VERSION
					] + $data,
					errors: $errors
				);
			} else throw new exception('Failed to initialize ' . static::TYPE . ' collection: ' . static::COLLECTION);
		} catch (exception $e) {
			// Writing to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}

		// Exit (fail)
		return null;
	}

	/**
	 * Read products
	 *
	 * @param string|null $search Search (text)
	 * @param string|null $filter Flter (AQL)
	 * @param string|null $sort Sort (AQL)
	 * @param int $page Page
	 * @param int $amount Amount per page
	 * @param string|null $return Return (AQL)
	 * @param string $language Language code (en, ru...)
	 * @param array &$errors Registry of errors
	 *
	 * @return array Массив с найденными товарами (может быть пустым)
	 */
	public static function read(
		?string $search = null,
		?string $filter = 'd.deleted != true && d.hidden != true',
		?string $sort = 'd.position ASC, d.created DESC',
		int $page = 1,
		int $amount = 100,
		?string $return = 'd',
		string $language = 'en',
		array &$errors = []
	): array {
		try {
			if (collection::initialize(static::COLLECTION, static::TYPE, errors: $errors)) {
				// Initialized the collection

				// Initializing the query
				$aql = <<<'AQL'
					FOR d IN @@collection
					AQL;

				if ($search) {
					// Requested search

					// Writing to the query
					$aql .=	<<<'AQL'
						 SEARCH 
							LEVENSHTEIN_MATCH(
								d.name.@language,
								TOKENS(@search, @analyzer)[0],
								1,
								false
							) OR
							levenshtein_match(
								d.description.@language,
								tokens(@search, @analyzer)[0],
								1,
								false
							) OR
							levenshtein_match(
								d.compatibility.@language,
								tokens(@search, @analyzer)[0],
								1,
								false
							) 
						AQL;

					// Adding sorting
					if ($sort) $sort = "BM25(d) DESC, $sort";
					else $sort = "BM25(d) DESC";
				}

				// Reading products
				$documents = collection::execute(
					sprintf(
						$aql . <<<'AQL'
						%s
						%s 
						LIMIT @offset, @amount
						RETURN $s
						AQL,
						empty($filter) ? '' : "FILTER $filter",
						empty($sort) ? '' : "SORT $sort",
						empty($return) ? 'd' : $return
					),
					[
						'@collection' =>	$search ? static::COLLECTION . 's_search' : static::COLLECTION,
						'search' => $search,
						'language' => $language,
						'analyzer' => "text_$language",
						'offset' => --$page <= 0 ? $page = 0 : $page * $amount,
						'amount' => $amount,
					],
					errors: $errors
				);

				if ($documents) {
					// Found products

					// Exit (success)
					return is_array($documents) ? $documents : [$documents];
				} else return [];
			} else throw new exception('Failed to initialize ' . static::COLLECTION . ' collection: ' . static::COLLECTION);
		} catch (exception $e) {
			// Writing to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}

		// Exit (fail)
		return [];
	}

	/**
	 * Collect parameter from all products
	 *
	 * @param string $name Name of the parameter (AQL path)
	 * @param array &$errors Registry of errors
	 *
	 * @return array Array with found unique parameter values from all products (can be empty)
	 */
	public static function collect(
		string $name = 'd._key',
		array &$errors = []
	): array {
		try {
			if (collection::initialize(static::COLLECTION, static::TYPE, errors: $errors)) {
				// Initialized the collection

				if ($values = collection::execute(
					<<<'AQL'
					FOR d IN @@collecton 
						RETURN DISTINCT @parameter
					AQL,
					[
						'@collection' => static::COLLECTION,
						'parameter' => $name
					],
					errors: $errors
				)) {
					// Found parameters

					// Exit (success)
					return $values;
				} else return [];
			} else throw new exception('Failed to initialize ' . static::TYPE . ' collection: ' . static::COLLECTION);
		} catch (exception $e) {
			// Writing to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}

		// Exit (fail)
		return [];
	}
}
